/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements. See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership. The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.gecko.rsa.discovery;

import java.util.Dictionary;
import java.util.Hashtable;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicReference;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.gecko.rsa.api.ExportEndpointHandler;
import org.osgi.framework.BundleContext;
import org.osgi.framework.Constants;
import org.osgi.framework.ServiceRegistration;
import org.osgi.service.remoteserviceadmin.EndpointDescription;
import org.osgi.service.remoteserviceadmin.EndpointEvent;
import org.osgi.service.remoteserviceadmin.EndpointEventListener;
import org.osgi.service.remoteserviceadmin.RemoteConstants;

/**
 * Listens for local EndpointEvents of services to be exported. This listener is usually called by 
 * the topology manager, that found some services to be exported. 
 * This will delegated to the {@link ExportEndpointHandler}, if available.
 */
public class ExportingEndpointListener implements EndpointEventListener {

    private static final Logger logger = Logger.getLogger(ExportingEndpointListener.class.getName());

    private final Map<String, Object> propertiesExtent = new ConcurrentHashMap<String, Object>();
    private final AtomicReference<ExportEndpointHandler> exportHandlerRef = new AtomicReference<ExportEndpointHandler>();
    private ServiceRegistration<EndpointEventListener> exportingRegistration;
	private String discoveryId;

    public ExportingEndpointListener(ExportEndpointHandler exportHandler, String discoveryId) {
    	this.exportHandlerRef.set(exportHandler);
		this.discoveryId = discoveryId;
    }
    
    /**
     * Returns a map where additional properties con be put in that are also used
     * for the service registration
     * @return
     */
    public Map<String, Object> getPropertiesExtent() {
    	return propertiesExtent;
    }
    
	/**
	 * Sets the exportHandler. This handler is used to delegate service description that will be
	 * exported to other frameworks.
	 * @param exportHandler the exportHandler to set
	 */
	public void setExportHandler(ExportEndpointHandler exportHandler) {
		this.exportHandlerRef.set(exportHandler);
	}
    
    /**
	 * Returns the discoveryId.
	 * @return the discoveryId
	 */
	public String getDiscoveryId() {
		return discoveryId;
	}
    
    /**
     * Registers the end-point listener for receiving changes on service descriptions that needs to be exported
     * @param bctx the bundle context
     */
    public void start(BundleContext bctx) {
    	if (discoveryId == null) {
    		throw new IllegalStateException("Cannot start the publishing endpoint listener, because repository or/and discoverId id/are null");
    	}
        Dictionary<String, Object> props = getRegistrationProperties(bctx);
        exportingRegistration = bctx.registerService(EndpointEventListener.class, this, props);
    }
    
	/**
     * Stops the endpoint listener for publishing endpoint descriptions
     */
    public void close() {
        if (exportingRegistration != null) {
            exportingRegistration.unregister();
            exportingRegistration = null;
        }
    }

    /* 
     * (non-Javadoc)
     * @see org.osgi.service.remoteserviceadmin.EndpointEventListener#endpointChanged(org.osgi.service.remoteserviceadmin.EndpointEvent, java.lang.String)
     */
    @Override
    public void endpointChanged(EndpointEvent event, String filter) {
        EndpointDescription endpoint = event.getEndpoint();
        switch (event.getType()) {
        case EndpointEvent.ADDED:
            addExported(endpoint, filter);
            break;
        case EndpointEvent.REMOVED:
            removeExported(endpoint, filter);
            break;
        case EndpointEvent.MODIFIED:
            modifyExported(endpoint, filter);
            break;
        }
    }
    
	/**
	 * Adds an end-point description, that has to be exported
	 * @param endpoint the end-point to be added
	 * @param filter the service filter
	 */
	void addExported(EndpointDescription endpoint, String filter)  {
		ExportEndpointHandler handler = exportHandlerRef.get();
		if (handler != null) {
			try {
				handler.exportServiceDescription(endpoint);
			} catch (Exception e) {
				logger.log(Level.SEVERE, String.format("[%s] Error delegating service description export to the export handler", endpoint.getId()), e);
			}
		} else {
			logger.info(String.format("[%s] Error delegating service description export to a null export handler", endpoint.getId()));
		}
	}

	/**
	 * Modifies an end-point description, that has to be exported
	 * @param endpoint the end-point to be modified
	 * @param filter the service filter
	 */
	void modifyExported(EndpointDescription endpoint, String filter) {
		ExportEndpointHandler handler = exportHandlerRef.get();
		if (handler != null) {
			try {
				handler.modifyServiceDescription(endpoint);
			} catch (Exception e) {
				logger.log(Level.SEVERE, String.format("[%s] Error delegating service description modificytion export to the export handler", endpoint.getId()), e);
			}
		} else {
			logger.info(String.format("[%s] Error delegating service description modification to a null export handler", endpoint.getId()));
		}
	}

	/**
	 * Removed an end-point description, that has to be exported
	 * @param endpoint the end-point to be removed
	 * @param filter the service filter
	 */
	void removeExported(EndpointDescription endpoint, String filter) {
		ExportEndpointHandler handler = exportHandlerRef.get();
		if (handler != null) {
			try {
				handler.removeServiceDescription(endpoint);
			} catch (Exception e) {
				logger.log(Level.SEVERE, String.format("[%s] Error delegating service description modificytion export to the export handler", endpoint.getId()), e);
			}
		} else {
			logger.info(String.format("[%s] Error delegating service description modification to a null export handler", endpoint.getId()));
		}
	}

	/**
	 * Returns the registration properties.
	 * @param bctx bundle context
	 * @return the properties dictionary
	 */
	private Dictionary<String, Object> getRegistrationProperties(BundleContext bctx) {
		Dictionary<String, Object> props = new Hashtable<String, Object>();
	    String uuid = bctx.getProperty(Constants.FRAMEWORK_UUID);
	    props.put(EndpointEventListener.ENDPOINT_LISTENER_SCOPE, 
	              String.format("(&(%s=*)(%s=%s))", Constants.OBJECTCLASS, 
	                            RemoteConstants.ENDPOINT_FRAMEWORK_UUID, uuid));
	    props.put(discoveryId, "true");
	    propertiesExtent.entrySet().forEach(e->props.put(e.getKey(), e.getValue()));
		return props;
	}
    
}
